#! /usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
import numpy         as np
from scipy.spatial   import KDTree, distance
from propobject      import BaseObject

from .tools          import is_arraylike

__all__ = ["get_hexprojection"]


def load_hexprojection(hexfile):
    """ """
    hex_ = HexagoneProjection(None, empty=True)
    hex_.load(hexfile)
    return hex_

def get_hexprojection(xy, ids=None, qdistance=None,
                      reference_ids=None, build=True, theta=None, **kwargs):
    """ """
    hgrid = HexagoneProjection(xy, ids=ids, qdistance=qdistance, **kwargs)
    
    reference_idx = hgrid.get_default_grid_reference() \
      if reference_ids is None else hgrid.ids_to_index(reference_ids)

    hgrid.set_grid_reference(*reference_idx, theta=theta)
    if build:
        hgrid.build_qr_grid(hgrid.ref_idx)
        
    return hgrid

class HexagoneProjection( BaseObject ):
    """ This class enable to build th Q,R hexagone grid
    based on given reference positions 
    """
    PROPERTIES         = ["xy","qdistance","hexgrid", "neighbors"]
    SIDE_PROPERTIES    = ["ref_idx","index_ids", "rot_degree"]
    DERIVED_PROPERTIES = ["hexafilled","grid_theta",
                              "tree","multipoint","rotmatrix",
                              "ids_index"]

    def __init__(self, xy, qdistance=None, load=True, ids=None,
                     empty=False):
        """ """
        if empty:
            return
        
        self.set_xy(xy, ids=ids)
        self.set_qdistance(qdistance)
        if load:
            self.fetch_neighbors()

    # -------------- #
    #   I/O          #
    # -------------- #
    def writeto(self, savefile):
        """ """
        from .tools import dump_pkl
        data = {
            "neighbors": self.neighbors,
            "ids":       self.index_ids,
            "hexgrid":   self.hexgrid,
            "ref_idx":   self.ref_idx,
            "qdistance": self.qdistance
            }
            
        dump_pkl(data, savefile)

    def load(self, hexfile):
        """ load the hexagone grid from the pkl file containing the data """
        from .tools import load_pkl
        data = load_pkl(hexfile)
        
        if "neighbors" not in data.keys():
            raise TypeError("The given filename does not have the appropriate format. No 'neighbors' entry.")

            
        # You can rerun everything with this
        self.set_neighbors(data["neighbors"])
        
        #  Valuable information
        if "ref_idx" in data.keys():
            self.set_grid_reference(*data["ref_idx"])

        #  No need to run anything with that again
        if "hexgrid" in data.keys(): #
            self.set_hexgrid(data["hexgrid"])
            
            
        if "ids" in data.keys():
            self.set_ids(data["ids"])

        if "qdistance" in data.keys():
            self.set_qdistance(data["qdistance"])


    def show(self, ax=None, switch_axis=False, colored_by=None, **kwargs):
        """ """
        from ..sedm import SEDMSPAXELS
        import matplotlib.pyplot as mpl
        from matplotlib      import patches
        if ax is None:
            fig = mpl.figure(figsize=[5,5])
            ax  = fig.add_subplot(111)
        else:
            fig = ax.figure

        indexes = list(self.ids_index.keys())
        if colored_by is None:
            colors = mpl.cm.viridis(np.random.uniform(size=len(indexes)))
        else:
            vmin, vmax = np.percentile(colored_by, [0,100])
            colors = mpl.cm.viridis( (colored_by-vmin)/(vmax-vmin) )
            
        ps = [patches.Polygon(SEDMSPAXELS + self.index_to_xy(self.ids_to_index(id_), switch_axis=switch_axis),
                           facecolor=colors[i], alpha=0.8) for i,id_  in enumerate(indexes)]

        ip = [ax.add_patch(p_) for p_ in ps]
        ax.autoscale(True, tight=True)
        
        fig.show()
    # -------------- #
    #   SETTER       #
    # -------------- # 
    def set_xy(self, xy, ids=None):
        """ """
        self._properties["xy"] = np.asarray(xy)
        self._derived_properties["tree"] = KDTree(self.xy)
        self._properties["hexgrid"] = None # reset
        self.set_ids(ids)
        
    def set_qdistance(self, qdistance=None):
        """ Distance at which the KDtree will be queried

        What should be the distance used to search the neighbors.
        If None, this optimal distance will be automatically built
        based on the density around the central point.
        If you are not sure, lease this to None
        """
        if qdistance is None:
            # We want 7 (the point  + 6 neiggtbors). So at what distance the 8th is?
            dist_8, index_8 = self.tree.query(self.xy[self.get_central_index()],8) 
            qdistance = np.mean(dist_8[-2:])
        
        self._properties["qdistance"] = qdistance

    def set_ids(self, ids):
        """ The id corresponding the given coordinates (xy). """
        self._side_properties["index_ids"] = ids
        
    def fetch_neighbors(self):
        """ """
        self.set_neighbors( self.tree.query_ball_tree(self.tree, r=self.qdistance) )

    def set_neighbors(self, neighbors):
        """ """
        self._properties["neighbors"] = neighbors
        
    def set_hexgrid(self, hexgrid):
        """ directly provide the (Q,R) coordinates for the indexes. """
        self._properties["hexgrid"] = hexgrid
        
    # ---------- #
    #   GETTER   #
    # ---------- #
    def get_central_index(self):
        """ the closest point to the centroid """
        if self.centroid is None:
            raise ValueError("This cannot estimate the centroid (no Shapely?).")
        
        return np.argmin([distance.euclidean(xydata_,self.centroid) for xydata_ in self.xy])

    def get_default_grid_reference(self):
        """ Look For the central value and 3 neightbors to define the 
        (0,0), (0,1) and (1,0) coordinates. 
        Returns
        -------
        3 indexes (with Q,R coords: (0,0), (0,1) and (1,0))
        """
        r00 = self.get_central_index()
        r01 = self.get_idx_neighbors(r00)[0]
        r10 = self.get_shared_neighbors(r00,r01)[0]
        return r00, r01, r10
    
    def get_idx_neighbors(self, index):
        """ Gives the name of all the `index` neightbors.
        There should be 6 (since it is an haxagonal grid) but less is expected
        if the index is at the edge of the detector. 
        More would be sign of problem

        Returns
        -------
        list of indexes
        """
        # No test here for fast recursive tricks
        arr = np.asarray(self.neighbors[index])
        return arr[arr!=index]

    def get_shared_neighbors(self, index1, index2):
        """ Get the indexes that are neighbors of both indexes """
        return np.intersect1d(self.get_idx_neighbors(index1),
                              self.get_idx_neighbors(index2),
                              assume_unique=True)

    def get_forth_index(self, n0, n1, ref):
        """ If n1 and n2 are neightbors, they should have two shared neighbors.
        One is ref the other one is returned.
        
        Returns
        -------
        index or None (if no forth index found)
        """
        index = [i for i in self.get_shared_neighbors(n0,n1) if i not in [ref]]
        if len(index)==0:
            return None
        if len(index)>2:
            warnings.warn("Several 'forth' indexes for n0 %d, n1 %d (ref %d)"%(n0,n1, ref))
            
        return index[0]
        
    # -------------- #
    #  Coordinates   #
    # -------------- #
    def index_to_ids(self, index):
        """ given the index of the given ids """
        return self.index_ids[index]

    def ids_to_index(self, ids):
        """ given the id of the given index """        
        if is_arraylike(ids):
            return np.asarray([self.ids_to_index(ids_) for ids_ in ids])
        
        return self.ids_index[ids]
    
    def index_to_qr(self, index):
        """ get the (Q,R) hexagonal coordinates of the given index"""
        return self.hexgrid[index]

    def index_to_xy(self, index, invert_rotation=True, switch_axis=False):
        """ """
        qr = np.asarray(self.index_to_qr(index)).T

        if not is_arraylike(index):
            q,r = qr if qr is not None else [np.NaN, np.NaN]
        else:
            q,r = np.asarray([qr_ if qr_ is not None else [np.NaN, np.NaN]
                                  for qr_ in qr]).T
            
        return self.qr_to_xy(q,r, invert_rotation=invert_rotation,
                            switch_axis=switch_axis)
    
    def qr_to_xy(self, q, r, invert_rotation=True,  switch_axis=False):
        """ Convert (q,r) hexagonal grid coordinates into (x,y) system 
        Returns
        -------
        x,y
        """
        # INVERTING X, Y on purpose to match the natural ifu_x ifu_y coordinates
        x,y = np.asarray([ (2*q + r)/np.sqrt(3.),  r])
        if invert_rotation:
           x,y = np.dot(self.grid_rotmatrix, np.asarray([x,y]))

        if switch_axis:
            x,y = y,x
            
        if self.rot_degree is not None and self.rot_degree != 0:
            _rot = self.rot_degree  * np.pi / 180
            rotmat = np.asarray([[ np.cos(_rot), -np.sin(_rot)],[ np.sin(_rot), np.cos(_rot)]])
            x,y = np.dot(rotmat, np.asarray([x,y]))
            
        return x,y
    
    # -------------- #
    # Grid Building  #
    # -------------- #
    def set_grid_reference(self, ref_00, ref_01, ref_10, theta=None):
        """  The three indexes defining (0,0),(0,1),(1,0).

        They must be neighbors.

        theta is the hexagonal grid angle (in radian)
        """
        if ref_10 not in self.get_shared_neighbors(ref_00, ref_01):
            raise ValueError("The given reference indexes are not neighbors.")

        self.hexgrid[ref_00] = [0,0]
        self.hexgrid[ref_01] = [0,1]
        self.hexgrid[ref_10] = [1,0]
        
        self._side_properties["ref_idx"] = [ref_00,ref_01,ref_10]
        self._derived_properties["hexafilled"] = None
        # - derived the baseline rotation matrix
        x,y = np.asarray(self.index_to_xy([ref_00,ref_01,ref_10], invert_rotation=False))
        if theta is None:
            v1 = np.asarray([x[1],y[1]])
            v2 = np.asarray([x[2],y[2]])
            theta = np.arccos(np.clip(np.dot(v1/np.linalg.norm(v1), v2/np.linalg.norm(v2)), -1.0, 1.0))

        self._derived_properties["grid_theta"] = theta
            
    def set_rot_degree(self, rot_degree):
        """ Rotation with respect to the north, in degree"""
        self._side_properties["rot_degree"] = rot_degree
        
    def populate(self, idx, neighbor0, neighbor1, ref):
        """ Based on the given neighbors and the reference index,
        the Q,R coordinates are set to the given idx.
        
        In the hexagonal grid, two neighbords N0 and N1 share
        two neighbors: M0 and M1. If you want to populate `idx=M0` then
        the `ref=M1`.

        The algorithm is as follows:
        M0_Q = (N0_Q - M1_Q) + N1_Q
        M0_R = (N0_R - M1_R) + N1_R.

        The Q,R coordinates are saved in self.hexgrid.
        
        If M1 (the reference) has no Q,R coordinates set yet, nothing happen. 0 returned
        If M0 (the index) already has Q,R coordinates, nothing happen. -1 returned
        If M0 (the index) has been populated, congrats, and 1 returned

        Returns
        -------
        1 : if index populated
        0 : if no Reference coords
        -1: if no need to population index
        (-2: if at least 1 neighbor is None)
        """
        
        # - No Reference
        if self.hexgrid[ref] is None:
            return 0
        if self.hexgrid[idx] is not None:
            return -1
        if self.hexgrid[neighbor0] is None or self.hexgrid[neighbor1] is None:
            return -2
        
        q_ref, r_ref = self.hexgrid[ref]
        q_0, r_0     = self.hexgrid[neighbor0]
        q_1, r_1     = self.hexgrid[neighbor1]
        self.hexgrid[idx] = [(q_0 - q_ref) + q_1, (r_0 - r_ref) + r_1]
        return 1


    def fetch_hexagon_references(self, hexcentral):
        """ """
        neightbors = self.get_idx_neighbors(hexcentral)
        for n1 in neightbors:
            if self.hexgrid[n1] is not None:
                n2 = [n2_ for n2_ in self.get_shared_neighbors(hexcentral, n1)
                          if self.hexgrid[n2_] is not None]
                if len(n2)>0:
                    return n1,n2[0]
        return None
        
    def populate_hexagon(self, central_hexagon):
        """ """
        refs = self.fetch_hexagon_references(central_hexagon)
        if refs is None:
            return
        n1_, ref1_ = refs
        index_to_populate = self.get_forth_index(central_hexagon, n1_, ref1_)
        
        # run to all. Passed it not necessary
        neighbors = self.get_idx_neighbors(central_hexagon)
        for i in neighbors:
            if index_to_populate is None:
                break
            out = self.populate(index_to_populate, central_hexagon, n1_, ref1_)
            n1_, ref1_        = index_to_populate, n1_
            index_to_populate = self.get_forth_index(central_hexagon, n1_, ref1_)
            
        self._hexafilled[central_hexagon] = True
        return neighbors

    def build_qr_grid(self, ref_idx):
        """ 
        Parameters
        ----------
        ref_idx: [3 ints]
            The three indexes defining (0,0),(0,1),(1,0).
            They must be neighbors. (see `set_grid_reference`)
        """
        if len(ref_idx) != 3:
            raise TypeError("The given ref_idx must be a list of 3 indexes")
        # - Defining the Axis
        self.set_grid_reference(*ref_idx)


        # - How will that be built?
        n0, n1, ref1 = self.ref_idx
        goahead      = True
        running_i    = 0
        next_to_run  = [n0]
        
        while(goahead and running_i<self.npoints):
            potential_next = self.populate_hexagon(n0)
            next_          = [n for n in potential_next if not self._hexafilled[n]] if potential_next is not None\
              else []                
            next_to_run    = [n_ for n_ in next_to_run if n_ not in [n0]] + next_
            if len(next_to_run)>1:
                n0 = np.random.choice(next_to_run)
                running_i +=1
            else:
                goahead = False


    # ----------- #
    #   PLOTTER   #
    # ----------- #

    # ================= #
    #   Properties      #
    # ================= #
    @property
    def xy(self):
        """ Data of the three """
        return self._properties["xy"]
    
    @property
    def npoints(self):
        """ """
        if self._properties["xy"] is not None:
            return len(self.xy)
        return len(self._properties["neighbors"]) if self._properties["neighbors"] is not None else None
        
    
    @property
    def qdistance(self):
        """ Distance used the query the neighbors """
        if self._properties["qdistance"] is None:
            self.set_qdistance(None)
        return self._properties["qdistance"]

    @property
    def index_ids(self):
        """ IDS corresponding the to indexes. If nothing set, this simply is this index. """
        if self._side_properties["index_ids"] is None:
            self._side_properties["index_ids"] = np.arange(self.npoints)
            
        return self._side_properties["index_ids"]

    @property
    def ids_index(self):
        """ IDs corresponding the to indexes. If nothing set, this simply is this index. """
        if self._derived_properties["ids_index"] is None:
            self._derived_properties["ids_index"] = {self.index_ids[i]:i for i in np.arange(self.npoints)}
            
        return self._derived_properties["ids_index"]


    # ----------
    # Side
    @property
    def rot_degree(self):
        """ """
        return self._side_properties["rot_degree"]
    # ----------
    # Derived
    @property
    def tree(self):
        """ The KDtree associated with the xy data"""
        return self._derived_properties["tree"]

    @property
    def neighbors(self):
        """ The neightbors """
        return self._properties["neighbors"]

    # - Building the Q,R grid
    @property
    def hexgrid(self):
        """ List containing the relation between index and (Q,R) """
        if self._properties["hexgrid"] is None:
            self._properties["hexgrid"] = np.asarray([None for i in range(self.npoints)])
        return self._properties["hexgrid"]

    @property
    def _hexafilled(self):
        """ Dictionary containing the relation between index and (Q,R)"""
        if self._derived_properties["hexafilled"] is None:
            self._derived_properties["hexafilled"] = {i:False for i in range(self.npoints)}
        return self._derived_properties["hexafilled"]

    
    @property
    def ref_idx(self):
        """ This indexes used to define (0,0), (0,1), (1,0). This is filled when running `set_grid_reference` """
        return self._side_properties["ref_idx"]

    @property
    def grid_theta(self):
        """ Angle [rad] that the (0,1), (0,0) (1,0) coordinates do. """
        if self._derived_properties["grid_theta"] is None:
            self._derived_properties["grid_theta"] = 0
        return self._derived_properties["grid_theta"] 
    
    @property
    def grid_rotmatrix(self):
        """ Rotation matrix associated to the grid. (based on self.grid_theta) """
        return np.asarray([[np.cos(self.grid_theta),np.sin(self.grid_theta)],
                            [-np.sin(self.grid_theta), np.cos(self.grid_theta)]])
    # ----------
    # Centroid 
    @property
    def centroid(self):
        """ """
        if self._multipoint is not None:
            return np.asarray(self._multipoint.centroid.xy).T[0]
        return None

    @property
    def _multipoint(self):
        """ """
        try:
            from shapely.geometry import MultiPoint, Point
        except ImportError:
            warnings.warn("No Shapely, no MultiPoint")
            return None
        
        if self._derived_properties["multipoint"] is None:
            self._derived_properties["multipoint"] = MultiPoint([Point(x,y) for x,y in self.xy])
            
        return self._derived_properties["multipoint"]



#! /usr/bin/env python
# -*- coding: utf-8 -*-


#################################
#
#   MAIN 
#
#################################
if __name__ == "__main__":
    
    import argparse
    import numpy as np
    import pyifu
    
    from pysedm import get_sedmcube, io

    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description=""" extract the A/B pair of a given cube""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('infile', type=str, default="None",
                        help='cube filepath')

    # // AUTOMATIC EXTRACTION
    parser.add_argument('--inputcube',   action="store_true", default=False,
                        help='Is the input a cube ?')
    
    parser.add_argument('--reducer',  type=str, default=io.SEDM_REDUCER,
                        help='What is your name? This will be added in the '
                             'header [default: $SEDM_REDUCER]')

    # Auto PSF
    parser.add_argument('--auto',  type=str, default=None,
                        help='Shall this run an automatic PSF extraction')

    parser.add_argument('--autorange',  type=str, default="4000,8000",
                        help='Wavelength range [in Angstrom] for measuring the '
                             'metaslice PSF')
    
    parser.add_argument('--autobins',  type=int, default=7,
                        help='Number of bins within the wavelength range '
                             '(see --autorange)')

    parser.add_argument('--buffer',  type=float, default=10,
                        help='Radius [in spaxels] of the aperture used for the '
                             'PSF fit. (see --centroid for aperture center)')

    parser.add_argument('--psfmodel',  type=str, default="NormalMoffatTilted",
                        help='PSF model used for the PSF fit: '
                             'NormalMoffat{Flat/Tilted/Curved}')

    parser.add_argument('--seeing',  type=float, default=2,
                        help='NOT READY YET Expected seeing (not quite in '
                             'arcsec). This helps the PSF model to converge.')

    # Centroids
    parser.add_argument('--centroidA',  type=str, default="auto", nargs="+",
                        help='Where is the A point source expected to be?'
                             '\nYou have 3 potential cases:'
                             '\n * --centroid d auto: Use astrometry from '
                             'guider if possible or falls back to the '
                             'brightest mode'
                             '\n * --centroid brightest: Use the position of '
                             'the bightest spaxels'
                             '\n * --centroid x0,y0: : Use the given x0 y0 '
                             'position (separated by a comma).')

    parser.add_argument('--centroidB', type=str, default="auto", nargs="+",
                        help='Where is the B point source expected to be?'
                             '\nYou have 3 potential cases:'
                             '\n * --centroid d auto: Use astrometry from '
                             'guider if possible or falls back to the '
                             'brightest mode'
                             '\n * --centroid brightest: Use the position of '
                             'the bightest spaxels'
                             '\n * --centroid x0,y0: : Use the given x0 y0 '
                             'position (separated by a comma).')
    
    parser.add_argument('--centroiderr',  type=str, default="None", nargs=2,
                        help="Where is the point source location guess "
                             "boundaries?"
                             "\nYou have 2 potential cases:"
                             "\n   * --centroiderr None: [3 3] if position "
                             "from the guider, [5 5] if from the brightest mode"
                             "\n   * --centroiderr [X Y]: use the given values")

    # display options
    parser.add_argument('--display',  action="store_true", default=False,
                        help='Select the area to fit using '
                             'the display function.')

    parser.add_argument('--vmin',  type=str, default="2",
                        help='Data Percentage used for imshow "vmin" '
                             'when using the --display mode')
    
    parser.add_argument('--vmax',  type=str, default="98",
                        help='Data Percentage used for imshow "vmax" '
                             'when using the --display mode')

    # Flux calibration
    parser.add_argument('--nofluxcal',  action="store_true", default=False,
                        help='No Flux calibration')
    
    parser.add_argument('--fluxcalsource',  type=str, default="None",
                        help="Path to a 'fluxcal' *.fits file. This file will "
                             "be used for the flux calibration"
                             "\nIf nothing given, the nearest (in time and "
                             "within the night) fluxcal will be used.")
    
    # wavelength step
    parser.add_argument('--lstep',  type=int, default=1,
                        help='Slice width in lbda step: default is 1, '
                             'use 2 for fainter source and maybe '
                             '3 for really faint target')

    # I/O options
    parser.add_argument('--tag',  type=str, default="None",
                        help='Add the tag on output filename.')

    parser.add_argument('--nofig',    action="store_true", default=False,
                        help='Do not create any plot')

    parser.add_argument('--notroncation',  action="store_true", default=False,
                        help='Do not apply the blue and red end edge '
                             'troncation on flux calibrated spectra')

    # - Standard Star object
    parser.add_argument('--std',  action="store_true", default=False,
                        help='Set this to True to tell the program you want to '
                             'build a calibration spectrum from this object')

    # Cosmic ray removal with byecr.py
    parser.add_argument('--byecr', action="store_true", default=False,
                         help="Shall this run byecr to remove cosmic rays at a specific spaxel at a specific wavelengh.")

    parser.add_argument('--byecr_lbda', type=float, default=None,
                         help="To check cosmic rays at a given wavelength.")

    parser.add_argument('--byecr_cut', type=float, default=5.0,
                         help="Cut criteria for byecr module.")

    parser.add_argument('--byecr_wspectral', action="store_true", default=False,
                         help="Also use a spectral filtering for byecr.")

    args = parser.parse_args()

    # ================= #
    #   The Scripts     #
    # ================= #

    # --------- #
    #  Date     #
    # --------- #
    if not args.inputcube:
        in_date = args.infile
    else:
        in_date = None

    # --------- #
    #  Parsing  #
    # --------- #
    if len(args.centroidA) == 1:
        args.centroidA = args.centroidA[0]
    elif len(args.centroidA) > 2:
        raise ValueError("--centroidA should be 'auto', 'brightest' or X Y")
    if len(args.centroidB) == 1:
        args.centroidB = args.centroidB[0]
    elif len(args.centroidB) > 2:
        raise ValueError("--centroidB should be 'auto', 'brightest' or X Y")

    # ------------- #
    #               #
    # 3D PSF        #
    #               #
    # ------------- #
    if args.auto is not None and len(args.auto) > 0:
        
        for target in args.auto.split(","):
            
            filecubes = io.get_night_files(
                in_date, "cube.*", target=target.replace(".fits", "")) \
                if not args.inputcube else [target]
            print("cube file from which the spectra will be extracted "
                  "[auto]: " + ", ".join(filecubes))

            # - loop over the file cube
            for filecube in filecubes:
                print("Automatic extraction of target %s, file: %s" %
                      (target, filecube))
                # ===================== #
                # Running ExtractStar   #
                # ===================== #
                #
                # DATA IN
                #
                cube = get_sedmcube(filecube)
                cube.header['REDUCER'] = args.reducer
                #
                # Extract B observation first
                #
                abdata = cube.data
                cube._derived_properties['data'] = 2000. - abdata
                #
                # OPTIONS
                #
                es_options = {
                    # Display mode ?
                    "display": args.display,
                    "displayprop": {"vmin": args.vmin, "vmax": args.vmax},
                    # Metaslices
                    # -- How many slices
                    "step1range": np.asarray(args.autorange.split(","),
                                             dtype="float"),
                    "step1bins": args.autobins,
                    # -- Centroid guess                    
                    "centroid": args.centroidB,
                    "prop_position": {"centroiderr": args.centroiderr},
                    # Sub IFU
                    "spaxelbuffer": args.buffer,
                    # PSF Model
                    "psfmodel": args.psfmodel,
                    "spaxels_to_use": None,  # not ready yet,
                    # provide here directly the spaxels to be used
                    "fwhm_guess": args.seeing,
                    # Spectral options
                    "slice_width": int(args.lstep),
                    #
                    "verbose": True,
                    }

                # Remove cosmic ray at a specific spaxel at a given wavelength.
                if args.byecr:
                    from pysedm import byecr

                    print(" Starting byecr cosmic ray removal ".center(50, "-"))
                    targetid = io.filename_to_id(filecube)
                    byecrclass = byecr.get_cr_spaxels_from_byecr(in_date, targetid)
                    cr_df = byecrclass.get_cr_spaxel_info(
                        lbda_index=args.byecr_lbda,
                        cut_criteria=args.byecr_cut,
                        wspectral=args.byecr_wspectral)

                    cube.data[cr_df["cr_lbda_index"], cr_df[
                        "cr_spaxel_index"]] = np.nan

                    print("Number of detected cosmic rays = %i" % len(cr_df))
                    print("Number of cosmic ray affected spaxels = %i" % len(
                        np.unique(cr_df["cr_spaxel_index"])))

                # ===================== #
                # SOURCE EXTRACTION     #
                # ===================== #
                print(" Starting B extract_pointsource ".center(50, "-"))
                try:
                    es_out = cube.extract_pointsource(**es_options)
                except IOError:
                    # Astrometry was not solved
                    print("ERROR: Astrometry file not found, "
                          "using brightest spaxel instead")
                    es_options["centroid"] = "brightest"
                    es_out = cube.extract_pointsource(**es_options)
                except AttributeError:
                    # Astrometry solved, but target outside IFU
                    print("ERROR: Astrometry places target outside IFU, "
                          "using brightest spaxel instead")
                    es_options["centroid"] = "brightest"
                    es_out = cube.extract_pointsource(**es_options)
                    cube.header['QUALITY'] = 3
                # -> The output object
                es_object_B = cube.extractstar
                spec_B = es_object_B.get_spectrum()
                #
                # Extract A observation next
                #
                cube._derived_properties['data'] = 2000. + abdata
                #
                # OPTIONS
                #
                es_options = {
                    # Display mode ?
                    "display": args.display,
                    "displayprop": {"vmin": args.vmin, "vmax": args.vmax},
                    # Metaslices
                    # -- How many slices
                    "step1range": np.asarray(args.autorange.split(","),
                                             dtype="float"),
                    "step1bins": args.autobins,
                    # -- Centroid guess
                    "centroid": args.centroidA,
                    "prop_position": {"centroiderr": args.centroiderr},
                    # Sub IFU
                    "spaxelbuffer": args.buffer,
                    # PSF Model
                    "psfmodel": args.psfmodel,
                    "spaxels_to_use": None,  # not ready yet,
                    # provide here directly the spaxels to be used
                    "fwhm_guess": args.seeing,
                    # Spectral options
                    "slice_width": int(args.lstep),
                    #
                    "verbose": True,
                }

                # ================= #
                # SOURCE EXTRACTION #
                # ================= #
                print(" Starting A extract_pointsource ".center(50, "-"))
                try:
                    es_out = cube.extract_pointsource(**es_options)
                except IOError:
                    # Astrometry was not solved
                    print("ERROR: Astrometry file not found, "
                          "using brightest spaxel instead")
                    es_options["centroid"] = "brightest"
                    es_out = cube.extract_pointsource(**es_options)
                except AttributeError:
                    # Astrometry solved, but target outside IFU
                    print("ERROR: Astrometry places target outside IFU, "
                          "using brightest spaxel instead")
                    es_options["centroid"] = "brightest"
                    es_out = cube.extract_pointsource(**es_options)
                    cube.header['QUALITY'] = 3
                # -> The output object
                es_object = cube.extractstar
                #
                # Generate output file tag
                add_tag = "_%s" % args.tag if args.tag is not None and \
                    args.tag not in ["None", ""] \
                    else ""
                add_info_spec = "_notfluxcal" if \
                    not es_object.is_spectrum_fluxcalibrated() else ""
                spec_info = "_lstep%s" % es_options["slice_width"] + \
                            add_info_spec

                print("Tag added", add_tag)
                # -
                # - PLOTTING: A is reference
                # -
                plot_tag = "_auto" + add_tag + spec_info + "_"
                if not args.nofig:
                    print(" Saving figures ".center(50, "-"))
                    es_object.show_adr(
                        savefile=es_object.basename.replace(
                            "{placeholder}", "adr_fit" + plot_tag))
                    es_object.show_metaslices(
                        savefile=es_object.basename.replace(
                            "{placeholder}", "metaslices" + plot_tag))
                    es_object.show_extracted_spec(
                        savefile=es_object.basename.replace(
                            "{placeholder}", "extracted_spec" + plot_tag))
                    es_object.show_mla(
                        savefile=es_object.basename.replace(
                            "{placeholder}", "ifu_spaxels_source" + plot_tag),
                        bcoords=(spec_B.header['XPOS'], spec_B.header['YPOS']))
                    es_object.show_psf(
                        savefile=es_object.basename.replace(
                            "{placeholder}", "psfprofile" + plot_tag),
                        sliceid=2)

                if args.byecr:
                    es_object.raw_spectrum.header.set("NCR", len(cr_df), "total number of detected cosmic-rays from byecr")
                    es_object.raw_spectrum.header.set("NCRSPX", len(np.unique(cr_df["cr_spaxel_index"])), "total number of cosmic-ray affected spaxels")

                # Generate A/B spectrum
                spec_A = es_object.get_spectrum()

                spec_AB = pyifu.get_spectrum(spec_A.lbda,
                                             (spec_A.data + spec_B.data) / 2.,
                                             variance=(spec_A.variance +
                                                       spec_B.variance) / 2.,
                                             header=spec_A.header)
                # Update A/B spectrum header
                spec_AB.header['EXPTIME'] = spec_A.header['EXPTIME'] + \
                                            spec_B.header['EXPTIME']
                spec_AB.header['ABPSFWA'] = spec_A.header['PSFFWHM']
                spec_AB.header['ABXPOSA'] = spec_A.header['XPOS']
                spec_AB.header['ABYPOSA'] = spec_A.header['YPOS']
                spec_AB.header['ABPSFWB'] = spec_B.header['PSFFWHM']
                spec_AB.header['ABXPOSB'] = spec_B.header['XPOS']
                spec_AB.header['ABYPOSB'] = spec_B.header['YPOS']
                es_object._derived_properties["spectrum"] = spec_AB
                # Save A/B spectrum
                es_object.writeto(basename=None, add_tag="auto" + add_tag,
                                  add_info=None)
                # Plot A/B spectrum
                es_object.spectrum.show(
                    savefile=es_object.basename.replace(
                        "{placeholder}", "spec" + plot_tag))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""  """
import warnings
import numpy as np
import matplotlib.pyplot as mpl

from astropy.io import fits as pf
from astropy.utils.console import ProgressBar


from propobject   import BaseObject
from .utils       import tools
from .sedm        import SEDM_CCD_SIZE

DEGREE   = 5
LEGENDRE = True

NGAUSS = 1
CONTDEGREE_GAUSS = 4
LEGENDRE_GAUSS= False


# ------------------ #
#  Builder           #
# ------------------ #
def build_background(ccd,
                    smoothing=[0,5],
                    start=2, jump=10, multiprocess=True, 
                    savefile=None, ncore=None):
    """ """
    from .io import is_stdstars, filename_to_background_name
    ccd.fit_background(start=start, jump=jump, multiprocess=multiprocess, 
                       set_it=False, is_std= is_stdstars(ccd.header), smoothing=smoothing,
                       ncore=ncore)
    
    ccd._background.writeto( filename_to_background_name(ccd.filename).replace('.gz','') )
    if savefile is not None:
        ccd._background.show(savefile=savefile)
        mpl.close("all")


# ------------------ #
#  Main Functions    #
# ------------------ #
def get_background(contvalues, size=SEDM_CCD_SIZE, smoothing=[0,5]):
    """ """
    back = Background()
    back.create(contvalues)
    back.build(size[0],size[1], smoothing=smoothing)
    return back

def load_background(filename):
    """ """
    back = Background()
    back.load(filename)
    return back

# ------------------- #
#   MultiProcessing   #
# ------------------- #
def get_contvalue(spec):
    """ """
    spec.fit_continuum(DEGREE, legendre=LEGENDRE)
    return spec.contmodel.fitvalues

def get_contvalue_sdt(spec):
    """ """
    spec.fit_continuum(CONTDEGREE_GAUSS, legendre=LEGENDRE_GAUSS, ngauss=NGAUSS)
    return spec.contmodel.fitvalues

def fit_background(ccd, start=2, jump=10, multiprocess=True,
                       ncore=None, is_std=False, show_progress=False):
    """ calling `get_contvalue` for each ccd column (xslice).
    This uses astropy's ProgressBar.map 

    Returns 
    -------
    dictionary 
    """
    # Running from ipython notebook
    
    
    index_column = range(ccd.width)[start::jump]
    if show_progress:
        notebook = tools.is_running_from_notebook()
        bar = ProgressBar( len(index_column), ipython_widget=notebook)
    else:
        bar = None

    if ncore is not None and ncore==1:
        if multiprocess:
            warnings.warn("you requested ncore=1, then multiprocess is set to False")
        multiprocess = False
        
    # - Multiprocessing 
    if multiprocess:
        import multiprocessing
        if ncore is None:
            if multiprocessing.cpu_count()>20:
                ncore = multiprocessing.cpu_count() - 10
            elif multiprocessing.cpu_count()>8:
                ncore = multiprocessing.cpu_count() - 5
            else:
                ncore = multiprocessing.cpu_count() - 2
            if ncore==0:
                ncore = 1
                
        p = multiprocessing.Pool(ncore)
        res = {}
        for j, result in enumerate( p.imap(get_contvalue if not is_std else get_contvalue_sdt, [ccd.get_xslice(i_) for i_ in index_column])):
            res[index_column[j]] = result
            if bar is not None:
                bar.update(j)
                
        if bar is not None:
            bar.update(len(index_column))
            
        return res
    
    # - No multiprocessing 
    return {index_column[i_]: get_contvalue(ccd.get_xslice(spec_)) if not is_std else get_contvalue_sdt(ccd.get_xslice(spec_)) for i_,spec_ in enumerate(bar)}

def _get_xaxis_polynomial_(xyv, degree=DEGREE, legendre=LEGENDRE,
                         xmodel=None, clipping = [5,5]):
    """ """
    from modefit.basics import get_polyfit
    x,y,v = xyv
    flagin = ((np.nanmean(y) - clipping[0] * np.nanstd(y)) < y) *  (y< (np.nanmean(y) + clipping[1] * np.nanstd(y)))
    
    contmodel = get_polyfit(x[flagin], y[flagin], v[flagin], degree=degree, legendre=legendre)
    contmodel.fit(a0_guess=np.nanmedian(y[flagin]))
    
    if xmodel is not None:
        return contmodel.fitvalues, contmodel.model.get_model(x=xmodel)#, contmodel
    return contmodel.fitvalues#, None, contmodel

        
class Background( BaseObject ):
    """ """
    PROPERTIES      = ["contvalues" ]
    SIDE_PROPERTIES = ['filename',"header","y"]
    DERIVED_PROPERTIES = ['input_columns', "input_background","background"]
    
    def load(self, filename):
        """ """
        data_ = pf.open(filename)
        
        # - background
        self._derived_properties['background'] = data_[0].data
        self._side_properties['header'] = data_[0].header
        # - contvalues
        contheader = data_["POLYVALUES"].header
        params = [int(k.replace("VALUE","")) for k,v in contheader.items() if "VALUE" in k]
        columns = data_["COLUMNS"].data
        convalues = {col:{contheader['VALUE%d'%i]: c_ for i,c_ in zip(params, cval_i)}
                         for col, cval_i in zip(columns, data_["POLYVALUES"].data)}
        self.create(convalues) 
        
    def create(self, contvalues):
        """ setup the instance based on the input 'contvalue' """
        self._properties['contvalues']    = contvalues
        self._derived_properties['input_columns'] = None
        
    def build(self, width, height, smoothing= [0,5]):
        """ """
        self._derived_properties['background'] = self.get_background(height, width, smoothing=smoothing)
        
    def contvalue_to_polynome(self, contvalue_):
        """ """
        from modefit.basics import polynomial_model, normal_and_polynomial_model
        if "mu0" in contvalue_.keys():
            poly = normal_and_polynomial_model( CONTDEGREE_GAUSS, NGAUSS)
            poly.use_legendre=LEGENDRE_GAUSS
        else:
            poly = polynomial_model( DEGREE )
            poly.use_legendre=LEGENDRE
            
        poly.setup([contvalue_[k] for k in poly.FREEPARAMETERS])
        self._side_properties["y"] = np.linspace(0, SEDM_CCD_SIZE[1], self.n_inputcolumns*2)
        model = poly.get_model(self._y)
        del(poly)
        return model


    def get_background(self, width, height, smoothing = [0,5]):
        """ """
        from scipy import ndimage, interpolate
        # get the blured image
        self._filtered = ndimage.filters.gaussian_filter( self.input_background, smoothing)
        # resample
        orig_shape = np.shape(self.input_background)
        x = np.linspace(0,1, orig_shape[0])
        y = np.linspace(0,1, orig_shape[1])
        nk = interpolate.RectBivariateSpline(x,y, self._filtered, kx=3,ky=3)
        
        return nk(np.linspace(0,1,width),np.linspace(0,1,height))


    # ----------------- #
    #   I/O and Plots   #
    # ----------------- #
    def writeto(self, savefile, overwrite=True, **header_kwargs):
        """ save the background as a .fits file
        
        The object will be structured as follows:
        0 PrimaryHDU: Background image 
        1 POLYVALUES: list of values associated to the best fits of column continuum
                      [parameter names in the header]
        2 COLUMNS: list of the fitted columns
        
        Use load() to read a .fits file created by this method.
        
        Parameters
        ----------
        savefile: [string]
            Full path of the .fits file to be created
            
        overwrite: [bool] -optional-
            Shall this overwrite an existing file if any.

        **header_kwargs additional information to be saved in the Primary header.
        
        Returns
        -------
        Void
        """
        img_shape = np.shape(self.background)
        
        self.header["NAXIS"] = len(img_shape)
        self.header['NAXIS1'] = img_shape[0]
        self.header['NAXIS2'] = img_shape[1]
        self.header['TYPE']   = "background"
        for k,v in header_kwargs:
            self.header[k] = v
            
        # --- Build the HDU
        hdu = [pf.PrimaryHDU(self.background, self.header)] # Background
        
        params = np.sort(list(list(self._contvalues.values())[0].keys()))
        header_POLY = pf.Header()
        for i, p in enumerate(params): 
            header_POLY['VALUE%s'%i] = p
            
        hdu.append(pf.ImageHDU( [[self._contvalues[i][p] for p in params] for i in self.input_columns], name='POLYVALUES', header=header_POLY))
        hdu.append(pf.ImageHDU(self.input_columns, name='COLUMNS'))
        
        hdulist = pf.HDUList(hdu)
        hdulist.writeto(savefile, overwrite =overwrite)

        
    def show(self, savefile=None, vmin=None, vmax=None, show=True, **kwargs):
        """ """
        from .utils.mpl import figout
        
        fig = mpl.figure(figsize=[9,4])
        space = 0.02
        width = 0.38
        axs  = fig.add_axes([0.1,0.1,width,0.8])
        ax   = fig.add_axes([0.1+width+space,0.1,width,0.8])
        axc  = fig.add_axes([0.1+(width+space)*2,0.1,0.02,0.8])
        if vmin is None:
            vmin = "2"
        if vmax is None:
            vmax = "98"
        prop = dict(origin="lower", aspect="auto", 
            vmax=np.percentile(self.input_background, float(vmax)) if type(vmax) == str else vmax,
            vmin=np.percentile(self.input_background, float(vmin)) if type(vmin) == str else vmin)
                          
        axs.imshow(self.input_background,        **prop)
        cl = ax.imshow(self.background, **prop)
        
        axs.set_title("column fit [source]")
        ax.set_title("background")
        ax.set_yticks([])
        fig.colorbar(cl, axc)
        
        fig.figout(savefile=savefile, show=show)
        
    # -------------------- #
    #    Properites        #
    # -------------------- #
    @property
    def n_inputcolumns(self):
        """ Number of column for which the background has been estimated"""
        return len(self._contvalues)

    @property
    def input_columns(self):
        """ index of the ccd column where the background has been estimated """
        if self._derived_properties['input_columns'] is None:
            self._derived_properties['input_columns'] = np.sort(list(self._contvalues.keys()))
            
        return self._derived_properties['input_columns']
    
    @property
    def _contvalues(self):
        """ """
        return self._properties['contvalues']

    @property
    def _y(self):
        """ """
        if self._side_properties['y'] is None:
            self._side_properties["y"] = np.linspace(0, 1, self.n_inputcolumns)
        return self._side_properties['y']
    

    # -- Background
    @property
    def background(self):
        """ """
        return self._derived_properties["background"]
    @property
    def input_background(self):
        """ """
        if self._derived_properties["input_background"] is None:
            self._derived_properties["input_background"] = \
              np.asarray([self.contvalue_to_polynome(self._contvalues[i])
                            for i in self.input_columns]).T
        return self._derived_properties["input_background"]
    
    # -- Fits tools
    @property
    def header(self):
        """ """
        if self._side_properties["header"] is None:
            self._side_properties["header"] = pf.Header()
        return self._side_properties["header"]
